"""
型タイプを定義する
"""
from typing import Any, Dict

# 型タイプ
Task = Dict[str, Any]
Inspection = Dict[str, Any]
Annotation = Dict[str, Any]
TaskHistory = Dict[str, Any]
TaskHistoryEvent = Dict[str, Any]
TaskId = str
InputDataId = str
